%{
#include "plasma.h"
#include "parser.h"

struct context {
	int			line;
	char			*filename;
	YY_BUFFER_STATE		state;
} con[32];

int level = -1;
int flushleft = 1;
int any_error = 0;
%}

%%
\'.*				/* ignore comment */
\".*\"				yytext[strlen(yytext) - 1] = 0; yylval.str = strdup(yytext + 1);	return TOK_STRING;
[0-9_]+				yylval.num = parseint(yytext, 10);					return TOK_INT;
\$[0-9a-fA-F_]+			yylval.num = parseint(yytext + 1, 16);					return TOK_INT;
0x[0-9a-fA-F_]+			yylval.num = parseint(yytext + 2, 16);					return TOK_INT;
\%\%[0-3_]+			yylval.num = parseint(yytext + 2, 4);					return TOK_INT;
0q[0-3_]+			yylval.num = parseint(yytext + 2, 4);					return TOK_INT;
\%[01_]+			yylval.num = parseint(yytext + 1, 2);					return TOK_INT;
0b[01_]+			yylval.num = parseint(yytext + 2, 2);					return TOK_INT;
"<<"													return TOK_SHL;
">>"													return TOK_SHR;
[#=()\[\]@~\-+*/^&|,]											return yytext[0];
:?[a-zA-Z_][a-zA-Z0-9_]*	yylval.str = strdup(yytext); flushleft = 0;				return TOK_ID;
[ \t\r]+			if(flushleft) 								return TOK_INDENT;
\n				con[level].line++; flushleft = 1;					return TOK_EOL;
<<EOF>>													return TOK_EOF;
.				yyerror(0, 0, "Unexpected character.");
%%

int yywrap() {
	if(level) {
		free(con[level].filename);
		yy_delete_buffer(con[level].state);
		level--;
		yy_switch_to_buffer(con[level].state);
		return 0;
	} else {
		return 1;
	}
}

static long long parseint(const char *str, int radix) {
	long long acc = 0;
	char c;

	while((c = *str++)) {
		if(c != '_') {
			if(c >= '0' && c <= '9') {
				c -= '0';
			} else if(c >= 'a' && c <= 'f') {
				c -= 'a' - 10;
			} else if(c >= 'A' && c <= 'F') {
				c -= 'A' - 10;
			}
			acc = acc * radix + c;
		}
	}

	return acc;
}

char *globalize(char *id, int declare) {
	static char global[128], context[128] = "0";
	if(id[0] == ':') {
		snprintf(global, sizeof(global), "%s%s", context, id);
		return global;
	} else if(declare) {
		snprintf(context, sizeof(context), "%s", id);
	}
	return id;
}

void yyerror(term_t p_label, term_t p_body, const char *msg) {
	fprintf(stderr, "%-16s%5d: %s\n", con[level].filename, con[level].line + 1, msg);
	any_error = 1;
}

static foreign_t visit(term_t filename) {
	char *fname;

	if(level == sizeof(con) / sizeof(*con) - 1) {
		fprintf(stderr, "Too many nested include directives.\n");
		exit(1);
	}
	if(PL_get_chars(filename, &fname, CVT_ATOM | CVT_STRING | CVT_LIST)) {
		FILE *fp = fopen(fname, "r");
		if(!fp) {
			PL_fail;
		}
		yyin = fp;
		level++;
		con[level].filename = strdup(fname);
		con[level].line = 0;
		con[level].state = yy_create_buffer(yyin, YY_BUF_SIZE);
		yy_switch_to_buffer(con[level].state);
		PL_succeed;
	}
	PL_fail;
}

static foreign_t parse(term_t label, term_t body) {
	return !yyparse(label, body);
}

static foreign_t complain(term_t message, term_t errflag) {
	char *msg;
	int flag;
	static int total = 0;

	if(PL_get_bool(errflag, &flag) && flag) any_error = 1;
	if(PL_get_chars(message, &msg, CVT_ATOM | CVT_STRING | CVT_LIST)) {
		fprintf(stderr, "%-16s%5d: %s\n", con[level].filename, con[level].line, msg);
		total++;
		if(total >= 40) {
			fprintf(stderr, "Too many errors.\n");
			exit(1);
		}
	}
	PL_succeed;
}

static foreign_t all_ok() {
	return !any_error;
}

static foreign_t location(term_t arg) {
	char *ftemp;

	if(PL_is_ground(arg)) {
		term_t file = PL_new_term_ref();
		term_t line = PL_new_term_ref();
		if(PL_get_arg(1, arg, file) && PL_get_arg(2, arg, line) && PL_get_chars(file, &ftemp, CVT_ATOM | CVT_STRING | CVT_LIST)) {
			free(con[level].filename);
			con[level].filename = strdup(ftemp);
			return PL_get_integer(line, &con[level].line);
		} else PL_fail;
	} else {
		return PL_unify_term(arg, PL_FUNCTOR_CHARS, "loc", 2, PL_CHARS, con[level].filename, PL_INT, con[level].line);
	}
}

static const PL_extension predtable[] = {
	{"visit",	1,	visit,		0},
	{"complain",	2,	complain,	0},
	{"parse",	2,	parse,		0},
	{"all_ok",	0,	all_ok,		0},
	{"location",	1,	location,	0},
	{0, 0, 0, 0}
};

install_t install() {
	PL_register_extensions(predtable);
}
