:- use_module(bits).

:- dynamic(symtab/4).

% State manipulation (state is: state(CogAddr, HubAddr))

 initial(state(0,   0  )).
     org(state(_,   Hub), state(Cog, Hub), cog, Cog).
     org(state(Cog, _  ), state(Cog, Hub), hub, Hub).
 address(state(Cog, Hub), Cog, Hub).
 padsize(state(Cog, _  ), N, 0)                        :- Cog mod N =:= 0, !.
 padsize(state(Cog, _  ), N, Pad)                      :- Pad is N - (Cog mod N).
     fit(state(Cog, _  ), Limit, cog, Verbose)         :- Regs is Cog >> 2, ignore(check(Regs, Limit, registers, Verbose)).
     fit(state(_,   Hub), Limit, hub, Verbose)         :- ignore(check(Hub, Limit, bytes, Verbose)).
 advance(state(InC, InH), state(OutC, OutH), Delta)    :- OutC is InC + Delta, OutH is InH + Delta.
     res(state(InC, Hub), state(OutC, Hub), cog, N)    :- OutC is InC + N * 4.
     res(state(Cog, InH), state(Cog, OutH), hub, N)    :- OutH is InH + N.

% Directives

directive(I,   I,    S,   S   ) --> [].
directive(I,   I,    S,   S   ) --> [include, Name],                      {visit(Name); err('Error opening "%p".', [Name])}.
directive(I,   I,    InS, OutS) --> [hres, Expr],                         {value(Expr, N), res(InS, OutS, hub, N)}.
directive(I,   I,    InS, OutS) --> [res, Expr],                          {value(Expr, N), res(InS, OutS, cog, N)}.
directive(I,   I,    InS, OutS) --> [horg, Expr],                         {value(Expr, Addr), org(InS, OutS, hub, Addr)}.
directive(I,   I,    InS, OutS) --> [org, Expr],                          {value(Expr, Addr), org(InS, OutS, cog, Addr)}.
directive(I,   I,    S,   S   ) --> [fit, Expr],                          {value(Expr, Limit), fit(S, Limit, cog, false)}.
directive(I,   I,    S,   S   ) --> [vfit, Expr],                         {value(Expr, Limit), fit(S, Limit, cog, true)}.
directive(I,   I,    S,   S   ) --> [hfit, Expr],                         {value(Expr, Limit), fit(S, Limit, hub, false)}.
directive(I,   I,    S,   S   ) --> [vhfit, Expr],                        {value(Expr, Limit), fit(S, Limit, hub, true)}.
directive(InI, OutI, InS, OutS) --> [pool],                               {buildpool(InI, CleanedI, Pool, InS, OutS), append(Pool, CleanedI, OutI)}.
directive(InI, OutI, InS, OutS) --> [long], datalist(4, InI, OutI, Size), {advance(InS, OutS, Size)}.
directive(InI, OutI, InS, OutS) --> [word], datalist(2, InI, OutI, Size), {advance(InS, OutS, Size)}.
directive(InI, OutI, InS, OutS) --> [byte], datalist(1, InI, OutI, Size), {advance(InS, OutS, Size)}.
directive(InI, OutI, InS, OutS) --> [align, Expr],                        {location(Loc), value(Expr, N), padsize(InS, N, Pad), buildlist(data(1, 0):Loc, Pad, InI, OutI), advance(InS, OutS, Pad)}.
directive(InI, OutI, InS, OutS) --> [file, Name],                         {location(Loc), exists_file(Name), size_file(Name, Size), OutI = [file(Name, Size):Loc | InI], advance(InS, OutS, Size)}.
directive(I,   I,    S,   S   ) --> [file, Name],                         {err('Error opening file: %p', [Name])}.
directive(InI, OutI, InS, OutS) --> code(Code),                           {location(Loc), OutI = [Code:Loc | InI], advance(InS, OutS, 4)}.

directive(I, I, S, S, _, []) :- err('Syntax error', []).

% Lists of data items

datalist(_, I, I, 0) --> [].

datalist(ItemSize, InI, OutI, Size) -->
	[Expr, repeat(NExpr)], !,
	{location(Loc), value(NExpr, N), buildlist(data(ItemSize, Expr):Loc, N, InI, NextI)},
	datalist(ItemSize, NextI, OutI, RestSize),
	{Size is N * ItemSize + RestSize}.

datalist(1, InI, OutI, Size) -->
	[Expr],
	{string(Expr), string_to_list(Expr, Bytes), appenddata(Bytes, InI, NextI)},
	!,
	datalist(1, NextI, OutI, RestSize),
	{string_length(Expr, Length), Size is Length + RestSize}.

datalist(ItemSize, InI, OutI, Size) -->
	[Expr],
	{location(Loc)},
	datalist(ItemSize, [data(ItemSize, Expr):Loc | InI], OutI, RestSize),
	{Size is ItemSize + RestSize}.

buildlist(_, 0, Tail, Tail) :- !.
buildlist(E, N, Tail, [E | Es]) :- Nm1 is N - 1, buildlist(E, Nm1, Tail, Es).

appenddata([], I, I).
appenddata([B | Bs], InI, OutI) :- location(Loc), appenddata(Bs, [data(1, B):Loc | InI], OutI).

% Pass 1 (parse tokens, declare symbols)

line(I, I, S, S, Id, [pool(Expr)]) :- atom(Id), !, declare(Id, Expr, Expr).	% pool here refers to the '=' in a constant definition
line(InI, OutI, InS, OutS, Label, Body) :-
	address(InS, CogAddr, HubAddr),
	CogReg is CogAddr >> 2,
	declare(Label, reg(CogReg), HubAddr),
	directive(InI, OutI, InS, OutS, Body, []).

pass1(InI, OutI, InS, OutS) :-
	parse(Label, Body),
	line(InI, NextI, InS, NextS, Label, Body),
	!,
	pass1(NextI, OutI, NextS, OutS).

pass1(InI, OutI, S, S) :- line(InI, OutI, S, _, [], [align, 4]).

% Pass 2 (evaluate expressions, generate code)

emit(Bytes, instr(Op, Flags, CC, Dest, imm(Src))) :- eval(Dest, D, false), eval(Src, S, false), putinstr(Op, Flags, 1, CC, D, S, Bytes).
emit(Bytes, instr(Op, Flags, CC, Dest, Src))      :- eval(Dest, D, false), eval(Src, S, false), putinstr(Op, Flags, 0, CC, D, S, Bytes).
emit(Bytes, pool(_, Expr))                        :- eval(Expr, N, false), checkpool(N), putlong(N, Bytes).
emit(Bytes, data(4, Expr))                        :- eval(Expr, N, false), putlong(N, Bytes).
emit(Bytes, data(2, Expr))                        :- eval(Expr, N, false), putword(N, Bytes).
emit([N],   data(1, Expr))                        :- eval(Expr, N, false).
emit(Bytes, file(Name, Size))                     :- open(Name, read, Stream, [type(binary)]), contents(Stream, Size, Bytes), close(Stream).

pass2([], []).
pass2([I:Loc | Is], OutBytes) :-
	pass2(Is, InBytes),
	location(Loc),
	emit(Bytes, I),
	!,
	append(Bytes, InBytes, OutBytes).

% Constant pool handling

buildpool([], [], [], S, S).
buildpool([instr(Op, Flags, CC, Dest, pool(Src)):Loc | InI], [instr(Op, Flags, CC, Dest, imm(Src)):Loc | OutI], Pool, InS, OutS) :-
	eval(Src, N, true),
	N < 512,
	!,
	buildpool(InI, OutI, Pool, InS, OutS).
buildpool([instr(Op, Flags, CC, Dest, pool(Src)):Loc | InI], [instr(Op, Flags, CC, Dest, reg(Reg)):Loc | OutI], OutPool, InS, OutS) :-
	!,
	buildpool(InI, OutI, InPool, InS, NextS),
	(member(pool(Reg, Src):_, InPool) ->
		OutPool = InPool,
		OutS = NextS
	;	address(NextS, CogAddr, _),
		Reg is CogAddr >> 2,
		OutPool = [pool(Reg, Src):Loc | InPool],
		advance(NextS, OutS, 4)).
buildpool([I | InI], [I | OutI], Pool, InS, OutS) :-
	buildpool(InI, OutI, Pool, InS, OutS).

checkpool(N) :- N >= 512; warn('Pool constant lower than 512. Impossible to know during first pass.', []).

% Evaluation of expressions

eval(reg(Res),          Res, _    ).
eval(Res,               Res, _    ) :- integer(Res).
eval(add(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 + E2.
eval(sub(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 - E2.
eval(mul(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 * E2.
eval(shl(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 << E2.
eval(shr(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 >> E2.
eval(and(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 /\ E2.
eval( or(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 \/ E2.
eval(xor(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet),           Res is E1 xor E2.
eval(div(Expr1, Expr2), Res, Quiet) :- eval(Expr1, E1, Quiet), eval(Expr2, E2, Quiet), E2 \== 0, Res is E1 // E2.
eval(not(Expr1),        Res, Quiet) :- eval(Expr1, E1, Quiet),                                   Res is \E1.
eval(neg(Expr1),        Res, Quiet) :- eval(Expr1, E1, Quiet),                                   Res is -E1.
eval(Sym,               Res, Quiet) :- atom(Sym), symtab(Sym, Expr, _, _), eval(Expr, Res, Quiet).
eval(hub(Sym),          Res, Quiet) :- atom(Sym), symtab(Sym, _, Expr, _), eval(Expr, Res, Quiet).
eval(Sym,               0,   false) :- atom(Sym), err('Undefined symbol: %p', [Sym]).
eval(hub(Sym),          0,   false) :- atom(Sym), err('Undefined symbol: %p', [Sym]).
eval(Expr,              0,   false) :- err('Invalid expression. (%p)', [Expr]).

% Symbol table

declare([], _, _).
declare(Id, CogExpr, HubExpr) :-
	symtab(Id, _, _, loc(File, Line)) ->
		err('%p already declared at line %p of %p.', [Id, Line, File])
	;	location(Loc),
		assert(symtab(Id, CogExpr, HubExpr, Loc)).

% Utility functions

err(Format, Params) :- swritef(Msg, Format, Params), complain(Msg, true).
warn(Format, Params) :- swritef(Msg, Format, Params), complain(Msg, false).

check(Value, Limit, Unit, _   ) :- Value > Limit, err('No fit! Using %p %p.', [Value, Unit]).
check(Value, Limit, Unit, true) :- all_ok, warn('Using %p %p out of %p.', [Value, Unit, Limit]).

putlong(N, [B3, B2, B1, B0]) :-
	B3 is (N >> 24) /\ 255,
	B2 is (N >> 16) /\ 255,
	B1 is (N >>  8) /\ 255,
	B0 is (N >>  0) /\ 255.

putword(N, [B1, B0]) :-
	B1 is (N >>  8) /\ 255,
	B0 is (N >>  0) /\ 255.

putinstr(Op, Flags, Imm, CC, D, S, Bytes) :-
	(S >= 512 -> warn('Source operand >= 512', []); true),
	(S < 0 -> warn('Source operand < 0', []); true),
	(D >= 512 -> warn('Destination operand >= 512', []); true),
	(D < 0 -> warn('Destination operand < 0', []); true),
	N is (Op << 26) \/ (Flags << 23) \/ (Imm << 22) \/ (CC << 18) \/ (D << 9) \/ S,
	putlong(N, Bytes).

contents(_, 0, []).
contents(Stream, N, [B | Bs]) :- Nm1 is N - 1, contents(Stream, Nm1, Bs), get_byte(Stream, B).

value(Expr, N) :- eval(Expr, N, false), !.

output(Bytes, Stream) :-
	reverse(Bytes, List),
	string_to_list(Code, List),
	write(Stream, Code).

% Main

options(_,       'native') --> [].
options(OutFile, Lib     ) --> ['-o', OutFile], options(_,       Lib).
options(OutFile, Lib     ) --> ['-l', Lib],     options(OutFile, _  ).

args(OutFile, InFile, Lib) --> options(OutFile, Lib), [InFile], {ground(OutFile)}.

main :-
	current_prolog_flag(argv, SysArgs),
	append(_, [-- | Args], SysArgs),
	(args(OutFile, InFile, Lib, Args, []) ->
		catch(load_foreign_library(Lib), _, (print('plasma: Failed to load the native library.'), nl, fail)),
		visit(InFile),
		initial(InS),
		pass1([], Instructions, InS, _),
		!, all_ok,
		%listing(symtab),
		pass2(Instructions, Bytes),
		!, all_ok,
		open(OutFile, write, Stream, [type(binary)]),
		output(Bytes, Stream),
		close(Stream)
	;	print('plasma: Invalid command line.'), nl, fail),
	halt.
